//Fichier Laplace.cpp modifie par Alain Perronnet
// a partir de l'original ecrit par Frederic HECHT  fem-v5/Laplace1.cpp
//------------------------------------------------------------------------------

//Resolution du Laplacien : Solution exacte U(x,y)=3*x+5*y

//------------------------------------------------------------------------------

int verbosity = 3;		//Definit le niveau des impressions

#include <cmath>
#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>		//pour lire le fichier de maillage

#include "rgraph.hpp"		//pour les traces
using namespace std;

#include "throwassert.hpp"
#include "MatriceCreuse_tpl.hpp"	//pour la matrice globale EF profil

#include "Quadrature.hpp"

using namespace Fem3D;

/*
 * Donnees du tenseur de conductivite
 * tenseurConductivite[i] est la diagonale du tenseur pour
 * le label "i"
 */
// hexaedre.msh :       2   {est le NUMERO dans VOLUME     de }HEXAEDRE                
#define MaxLabel 6
R tenseurConductivite[][3] = {
	{1., 1., 1.},		// idendite, le plus classique
	{1., 1., 1.},		// idendite, le plus classique
	{1., 1., 1.},		// idendite, le plus classique
	{1., 1., 1.},		// idendite, le plus classique
	{2., .5, .5},		// surtout sur l'axe des x
	{.5, 2., .5},		// surtout sur l'axe des y
	{.5, .5, 2.},		// surtout sur l'axe des z
};

//------------------------------------------------------------------------------
//Construction de la matrice elementaire Ae sur l'element K

void GradGrad(MatriceElementaireSymetrique < R > &Ae, const Tetraedre & tetra, int, int, void *)	//3 parametres inutilises ici ie,label,stack
	       //(GradGrad est un pointeur de fonction)
{
	R *aa = Ae.a;		//pointeur sur le coefficient courant de la matrice elementaire

//  cout << endl << "Grad grad : "<< endl;
	for (int i = 0; i < 4; i++) {
		R3 Hi = tetra.H(i);
		int l = ((Label) tetra).lab;
		assert(l == 2);	// for hexaedre.msh
		assert(l < MaxLabel);	// check for tenseurConductivite[][]
		R3 gmi(Hi.x * tenseurConductivite[l][0],	// gradient modifie = A(x)*grad(wi)
		       Hi.y * tenseurConductivite[l][1],	// ou A(x) est diagonale
		       Hi.z * tenseurConductivite[l][3]);
		for (int j = 0; j <= i; j++) {	// grad.grad est constant egale a (A.Hi,Hj)
			*aa++ = (gmi, tetra.H(j)) * tetra.volume;	// sur le tetraedre -> *= volume
//      cout << "   Hi,Hj and whole : ("<< Hi<<  "), (" << tetra.H( j ) << ") and " << ( ( gmi,tetra.H( j ) ) * tetra.volume ) << endl;
		}
	}
	assert((aa - Ae.a) == 10);
}

//------------------------------------------------------------------------------
//


// pour l'hexaedre : 
R3 normale(Label l)
{
	switch (l.lab) {
	case 1:
		return R3(0, 0, -1);
	case 2:
		return R3(0, 0, +1);
	case 3:
		return R3(-1, 0, 0);
	case 4:
		return R3(+1, 0, 0);
	case 5:
		return R3(0, -1, 0);
	case 6:
		return R3(0, +1, 0);
	default:
		assert(false);	// should not happen 
		return R3(0, 0, 0);	// prevent a warnign from gcc

	}
}



/*
 * les donnees
 */


//#include "donnees_dirichlet_uniforme.cpp"
//#include "donnees_dirichletnul.cpp"
#include "donnees_quelconque.cpp"
//#include "donnees_x2y2z2.cpp"



//------------------------------------------------------------------------------
//Le programme principal

int main(int argc, char **argv)
{
	cout << "Debut du programme" << endl;

	Mesh Th((argc < 2) ? MESHFILE : argv[1]);
	Th.renum();		//renumerotation des noeuds pour reduire le profil de la matrice

	cout << "Mesh charge" << endl;

	MatriceProfile < R > Ag(Th);	//Constructeur de la matrice globale
	MatriceElementaireSymetrique < R > Ae(Th);	//Constructeur matrice elementaire

	cout << "Matrices allouees" << endl;

	Ae.element = GradGrad;
	for (int i = 0; i < Th.ntetra; i++)
		Ag += Ae(i, -1, 0, 0);
#if 0
	cout << "-----------------  Dumping Ag : begin ------------------"
	    << endl;
	cout << "-----------------  (just after Ag += Ae(i,-1,0,0))------"
	    << endl;
	for (int i = 0; i < Ag.pL[Ag.n]; i++)
		cout << "L[" << i << "]= " << Ag.L[i] << endl;
	cout << "-----------------  Dumping Ag : end  ------------------"
	    << endl;
#endif

	cout << "Matrice Ag assemblee" << endl;

	RN Ug(Th.nv);		//Tableau de taille le nombre de vertices
	Ug = 0;			//Mise a zero

	/*
	 * Calcul du second membre global
	 */
	for (int i = 0; i < Th.ntetra; i++) {
		Tetraedre & tetra = Th[i];
		R volume = tetra.volume;

		// get vertices indices
		int i0 = Th(i, 0);
		assert(i0 >= 0);
		assert(i0 < Th.nv);
		int i1 = Th(i, 1);
		assert(i1 >= 0);
		assert(i1 < Th.nv);
		int i2 = Th(i, 2);
		assert(i2 >= 0);
		assert(i2 < Th.nv);
		int i3 = Th(i, 3);
		assert(i3 >= 0);
		assert(i3 < Th.nv);

		// choix de la formule d'integration
		Quadrature3DPoint *QF = Formula3D_4Points;
		int QFN = 4;
		//Quadrature3DPoint  *QF =  Formula3D_15Points; int QFN = 15;

		for (int j = 0; j < QFN; j++)	//loop on integration's points
		{
			Quadrature3DPoint & PI = QF[j];	//XY of the integration point j 
			R w1(PI.x), w2(PI.y), w3(PI.z), w0(1 - w1 - w2 - w3);	// Coordonnees barycentriques
			assert(w0 >= 0.);	// paranoiac check

			R3 P = tetra[0] * w0 + tetra[1] * w1 + tetra[2] * w2 + tetra[3] * w3;	// point courant

			R F = fOmega(P) * PI.poids * volume * 6.;

			Ug[i0] += F * w0;	//vecteur global partie interne
			Ug[i1] += F * w1;	//assemble a partir du vecteur elementaire
			Ug[i2] += F * w2;
			Ug[i3] += F * w3;
		}
	}

	cout << "Fin du calcul du second membre" << endl;

	/* 
	 * Frontiere de Neumann
	 */
#if 1
	for (int i = 0; i < Th.nbt; i++)	// boucle sur les boundary triangles
	{
		Triangle & T = Th.getBoundaryTriangle(i);
		R area = T.area;
		Label label = ((Label) T).lab;

		if (!isNeumann(label))
			continue;

		int i0 = Th.getBoundaryTriangleVerticeIndice(i, 0);
		int i1 = Th.getBoundaryTriangleVerticeIndice(i, 1);
		int i2 = Th.getBoundaryTriangleVerticeIndice(i, 2);
		assert(i0 >= 0);
		assert(i0 < Th.nv);
		assert(i1 >= 0);
		assert(i1 < Th.nv);
		assert(i2 >= 0);
		assert(i2 < Th.nv);

		Quadrature2DPoint *QF = Formula2D_3PointsAxisymetrie;
		int QFN = 3;
		//Quadrature2DPoint  *QF = Formula2D_7Points; int QFN=7;

		for (int j = 0; j < QFN; j++)	//loop on integration's points
		{
			Quadrature2DPoint & PI = QF[j];

			R w1(PI.x), w2(PI.y), w0(1 - PI.x - PI.y);

			R3 P =
			    (R3) T[0] * w0 + (R3) T[1] * w1 +
			    (R3) T[2] * w2;

			R N = 2. * area * UNeumann(P, label) * PI.poids;	// contribution de ce point

			Ug[i0] += N * w0;
			Ug[i1] += N * w1;
			Ug[i2] += N * w2;
		}
	}
	cout << "Fin du calcul pour les conditions de Neumann" << endl;
#endif


	/* 
	 * Frontiere de Robin
	 */
#if 1
	for (int i = 0; i < Th.nbt; i++)	// boucle sur les boundary triangles
	{
		Triangle & T = Th.getBoundaryTriangle(i);
		R area = T.area;
		Label label = ((Label) T).lab;

		if (!isRobin(label))
			continue;

		int i0 = Th.getBoundaryTriangleVerticeIndice(i, 0);
		int i1 = Th.getBoundaryTriangleVerticeIndice(i, 1);
		int i2 = Th.getBoundaryTriangleVerticeIndice(i, 2);
		assert(i0 >= 0);
		assert(i0 < Th.nv);
		assert(i1 >= 0);
		assert(i1 < Th.nv);
		assert(i2 >= 0);
		assert(i2 < Th.nv);

		Ae.m = Ae.n = 3;
		Ae.ni[0] = i0;
		Ae.ni[1] = i1;
		Ae.ni[2] = i2;
		for (int i = 0; i < 3 * 4 / 2; i++)
			Ae.a[i] = 0.;

		Quadrature2DPoint *QF = Formula2D_3PointsAxisymetrie;
		int QFN = 3;
		//Quadrature2DPoint  *QF = Formula2D_7Points; int QFN=7;

		for (int j = 0; j < QFN; j++)	//loop on integration's points
		{
			Quadrature2DPoint & PI = QF[j];

			R w1(PI.x), w2(PI.y), w0(1 - PI.x - PI.y);
			R3 P =
			    (R3) T[0] * w0 + (R3) T[1] * w1 +
			    (R3) T[2] * w2;

			R N = GRobin(P, label) * 2 * area * PI.poids;

			/* contribution a la matrice */
			Ae(0, 0) += N * w0 * w0;
			Ae(0, 1) += N * w0 * w1;
			Ae(0, 2) += N * w0 * w2;
			Ae(1, 1) += N * w1 * w1;
			Ae(1, 2) += N * w1 * w2;
			Ae(2, 2) += N * w2 * w2;

			/* contribution au second membre */
			N *= TetaExtRobin(P);
			Ug[i0] += N * w0;
			Ug[i1] += N * w1;
			Ug[i2] += N * w2;
		}
		Ag += Ae;
	}

	cout << "Fin du calcul pour les conditions de Robin/Fourier" <<
	    endl;
#endif

	/* 
	 * Frontiere de Dirichlet
	 */
	const R tgv = 1e30;
	for (int i = 0; i < Th.nv; i++)	//Traitement des noeuds sur la frontiere Dirichlet
		if (isDirichlet(((Label) Th(i)).lab))
			Ug[i] = UDirichlet(Th(i)) * (Ag.D[i] = tgv);

	cout << "Fin du calcul pour les conditions de Dirichlet" << endl;


	Ag.cholesky();		// Ag = L tL   Factorisation de la matrice globale apres C.L.
#if 0
	cout << "-----------------  Dumping Ag : begin ------------------"
	    << endl;
	cout << "-----------------  (just after Ag.cholesky()) ----------"
	    << endl;
	for (int i = 0; i < Ag.pL[Ag.n]; i++)
		cout << "L[" << i << "]= " << Ag.L[i] << endl;
	cout << "-----------------  Dumping Ag : end  ------------------"
	    << endl;
#endif


	Ug /= Ag;		// Ug = Ag**-1 Ug Resolution du systeme lineaire

	cout << "Fin du calcul de l'inverse de la matrice" << endl;


#if 0
	cout << endl << "Comparaison avec tetareel << " << endl;
	for (int i = 300; i < 360; i++)
		if (!isDirichlet(Th.vertices[i]))
			cout << "  Ug reel,Ug[" << i << "] are " <<
			    TetaExact(Th.
				      vertices[i]) << "," << Ug[i] <<
			    "  (Ugreel-Ug)[] is " <<
			    (TetaExact(Th.vertices[i]) -
			     Ug[i]) << " as %% :" <<
			    float ((TetaExact(Th.vertices[i]) -
				    Ug[i])) * 100. /
	    float (TetaExact(Th.vertices[i])) << endl;
#endif


#if 0
	//Trace graphique de la temperature sous forme d'isothermes
	initgraphique();
	RN Viso(20);
	Th.InitDraw();
//  SetDefaultIsoValue( Ug,Viso);
	reffecran();
//  openPS("laplace");
//  PlotValue(Viso,0,"Isovalue");
//  Vh.Draw(Ug,Viso);
	closePS();
//  rattente(1);
	closegraphique();
#endif

	// test de validite
	R max = 0.;
	R tetamax = 0.;
	R tetamin = 0.;
	R tetamax_real = 0.;
	R tetamin_real = 0.;
	int under, toomuch;
	under = toomuch = 0;
	assert(Ug.N() == Th.nv);
	for (int i = 0; i < Th.nv; i++)	// on parcours l'ensemble des points
	{
		R diff;
		Vertex V;
		int l;

		V = Th.vertices[i];
		l = ((Label) V).lab;
		diff = Ug[i] - TetaExact(V);
		diff *= diff;	// on prends le carre
//        if ( !isDirichlet( l )) cout << "valeur trouvee/exacte : " << Ug[ i ] << ", " << TetaExact(V)<< " diff : " << diff << " label : "<<l << endl;
		if (abs(TetaExact(V)) > .001)	// else dont calculate
			if ((diff / (TetaExact(V) * TetaExact(V))) > 0.01) {
				toomuch++;
//                        cout << "un toomuch en " << ( R3 )V << endl;
			} else {
				under++;	// 2% tolerated
//                        if ( V.x *V.y *V.z * ( 1-V.x )*(  1-V.y )*(  1-V.z )) cout << "un under en " << ( R3 )V << endl;
			}
		if (diff > max)
			max = diff;
		if (Ug[i] > tetamax)
			tetamax = Ug[i];
		if (Ug[i] < tetamin)
			tetamin = Ug[i];
		if (TetaExact(V) > tetamax_real)
			tetamax_real = TetaExact(V);
		if (TetaExact(V) < tetamin_real)
			tetamin_real = TetaExact(V);
	}
	printf("\nMax des carres obtenu : %f", max);
	printf("\n               sqrt   : %f", sqrt(max));
	printf("\nMin,Max des tetas obtenu : %f, %f", tetamin, tetamax);
	printf("\nMin,Max des tetas reels  : %f, %f", tetamin_real,
	       tetamax_real);
	printf("\nStats under,toomuch   : %d, %d", under, toomuch);
	printf("\nTaux d'erreur : %f\n\n",
	       float (sqrt(max) * 100. /
		      float (tetamax_real - tetamin_real)));



	return 0;
}
