#ifndef  MatriceCreuse_tpl_
#define  MatriceCreuse_tpl_
#ifndef  MatriceCreuse_h_
#include "MatriceCreuse.hpp"
#endif


template<class R>
void MatriceElementaireSymetrique<R>::call(int k,int ie,int label,void * stack)
{

  assert ( k>=0 ); assert (k< Th.ntetra); assert(element);

  const Tetraedre &tetra = Th[k];
  int nbdf = 4;


  // copy the numbering 
  m=n = nbdf; 
  for (int i=0;i<nbdf;i++)
     ni[i] = Th(k,i);

  // mise a zero de la matrice elementaire, plus sur
  int n2 = ((nbdf+1)*nbdf)/2;
  for (int i=0;i<n2;i++) a[i]=0.;

  // call element to actually constructs the matrix
  element(*this,tetra,ie,label,stack); 

}


template<class R>
MatriceProfile<R>::~MatriceProfile() {
  if(!dummy) 
    { //cout << " del mat profile " << endl ;
    if (U && (U !=L))  delete [] U;
    if (D)  delete [] D;
    if (L)  delete [] L;
    if (pU && (pU != pL)) delete [] pU;
    if (pL) delete [] pL;
    //cout << " dl de MatriceProfile " << this << endl;
    }
}


template<class R>
int MatriceProfile<R>::size() const {
  int s = sizeof(MatriceProfile<R>);
  if (D) s += n*sizeof(R);
  if (pL) s += n*sizeof(int);
  if (pU && (pU != pL)) s += n*sizeof(int);
  if (L) s += pL[n]*sizeof(int);
  if (U && (U != L)) s += pU[n]*sizeof(int);
  return s;
}


/*
template<class R>
MatriceProfile<R>::MatriceProfile(const int nn,const R *a)
  :MatriceCreuse<R>(nn,nn,0),typefac(FactorizationNO)
{
  int *pf = new int [n+1];
  int i,j,k;
  k=0;
  for (i=0;i<=n;k+=i++)
    {
      pf[i]=k;
      //  cout << " pf " << i<< " = " << k  << endl;
    }
  throwassert( pf[n]*2 == n*(n-1));
  pU = pf; // pointeur profile U
  pL = pf; // pointeur profile L
  U = new R[pf[n]];
  L = new R[pf[n]];
  D = new R[n];  
  const R *aij=a;
  for (i=0;i<n;i++)
    for (j=0;j<n;j++)
      if      (j<i)   L[pL[i+1]-i+j] = *aij++;
      else if (j>i)   U[pU[j+1]-j+i] = *aij++;
      else            D[i] = *aij++;
}
*/


template<class R>
MatriceProfile<R>::MatriceProfile(const Mesh & Th) 
  :MatriceCreuse<R>(Th.nv,Th.nv,0),typefac(FactorizationNO) // 4 is the # of Degrees of Freedom
{
  dummy=0;
  n = m = Th.nv;
  int i,j,k,mn,jl;

  // creation / initialisation du profil
  int *pf = new int [n+1];
  for (i=0;i<n;i++)  pf[i]=0;


  for (k=0;k<Th.ntetra;k++) { 
      mn = n;
      for (jl=0;jl<4;jl++)
	mn = Min ( mn , Th( k,jl ) ) ;
      // mn contient le plus petit indice du triangle k
      for (jl=0;jl<4;jl++) {
	  j = Th( k,jl );
	  pf[j] = Max(pf[j],j-mn);
	  }
      }
  /*----------------------------------------------------------------
    D[i] = A[ii]
    L[k] = A[ij]  j < i avec:   pL[i]<= k < pL[i+1] et j = pL[i+1]-k
    U[k] = A[ij]  i < j avec:   pU[j]<= k < pU[j+1] et i = pU[i+1]-k 
    remarque pL = pU generalement 
    si L = U => la matrice est symetrique 

    quand on a i,j :
	if      (j<i)  L[ pL[i+1] - (i-j) ] = XXXXXX
	else if (j>i)  U[ pU[j+1] - (j-i) ] = XXXXXX
	else -> Diagonale D[ i ]
    -------------------------------------------------------------------
  */ 

  int l =0;
  for (i=0;i<n;i++)  {int tmp=l;l += pf[i]; pf[i]=tmp;}

  pf[n] = l;
  if(verbosity >3) 
    cout << "  -- SizeOfSkyline =" <<l << endl;

  pU = pf; // pointeur profile U
  pL = pf; // pointeur profile L
  D = 0; // diagonal
  U = 0; // upper part
  L = 0; // lower part 
}


template<class R>
void MatriceProfile<R>::addMatMul(const KN_<R> &x,KN_<R> &ax) const 
{if (x.n!= n ) ERREUR(MatriceProfile MatMut(xa,x) ," longueur incompatible x (in) ") ;
 if (ax.n!= n ) ERREUR(MatriceProfile MatMut(xa,x) ," longueur incompatible ax (out)") ;
 int i,j,k,kf;
 throwassert(n == m);
 if (D) 
   for (i=0;i<n;i++) 
     ax[i] += D[i]*x[i];
 else
   for (i=0;i<n;i++) // no dia => identyty dai
     ax[i] +=x[i];
      
 if (L && pL )    
   for (kf=pL[0],i=0;  i<n;   i++  )  
     for ( k=kf,kf=pL[i+1], j=i-kf+k;   k<kf; j++,  k++  )
       ax[i] += L[k]*x[j],throwassert(i>=0 && i <n && j >=0 && j < m && k>=0 && k < pL[n]);
       
 if (U && pU)     
   for (kf=pU[0],j=0;  j<m;  j++)  
     for (k=kf,kf=pU[j+1], i=j-kf+k;   k<kf; i++,  k++  )
       ax[i] += U[k]*x[j],throwassert(i>=0 && i <n && j >=0 && j < m &&  k>=0 && k < pU[n]);
}


template<class R>
void MatriceProfile<R>::operator=(const R & v) {
  if(v!=0.0)
    { cerr << " Mise a zero d'une matrice MatriceProfile<R>::operator=(R v) uniquement v=" << v << endl;
    throw(ErrorExec("exit",1));
    }
  typefac = FactorizationNO;
  delete [] U;
  delete [] L;
  delete [] D;
  U=L=D=0;
}
template<class R>
MatriceCreuse<R>  & MatriceProfile<R>::operator +=(MatriceElementaire<R> & me) {
  int il,jl,i,j,k;
  int * mi=me.ni, *mj=me.nj;

  assert( me.mtype==MatriceElementaire<R>::Symetric ); // others still untested

  if (!D)  // matrice vide 
    {
    D  = new R[n]; for (i =0;i<n;i++) D[i] =0.;
    L  = pL[n] ? new R[pL[n]] :0 ;
    for (k =0;k<pL[n];k++) L[k] =0.;
    switch (me.mtype) {
    case MatriceElementaire<R>::Full :     
      U  = pU[n] ? new R[pU[n]] : 0;
      for (k =0;k<pU[n];k++) U[k] =0;
      break;
    case MatriceElementaire<R>::Symetric :     
      U = L; 
      break;
    default:
      cerr << "Big bug type MatriceElementaire unknown" << (int) me.mtype << endl;
      throw(ErrorExec("exit",1));
      break; 
    } // switch
    }// if
  R * al = me.a; 
//  cout << "operator *= : me.n, me.m are " << me.n << " " << me.m << endl;
  switch (me.mtype) {
  case MatriceElementaire<R>::Full : //throwassert(L !=U);
    for (i=mi[il=0]; il<me.n; i=mi[++il])  
      for ( j=mj[jl=0]; jl< me.m ; j=mj[++jl],al++)  
	if      (j<i)  L[ pL[i+1] - (i-j) ] += *al;
	else if (j>i)  U[ pU[j+1] - (j-i) ] += *al;
	else           D[i] += *al;
    break;
     
  case MatriceElementaire<R>::Symetric : //throwassert(L ==U);   
    for (i=mi[il=0]; il<me.n; i=mi[++il])  
      for (j=mj[jl=0];jl< il+1 ; j=mj[++jl])  
	if      (j<i)  L[ pL[i+1] - (i-j) ] += *al++;
	else if (j>i)  U[ pU[j+1] - (j-i) ] += *al++;
	else           D[i] += *al++;
    break;
  default:
    cerr << "Big bug type MatriceElementaire unknown" << (int) me.mtype << endl;
    exit(1);
    break; 
  }      
  return *this;
} 


template<class R>
ostream& MatriceProfile<R>::dump (ostream& f) const 
{f<< " matrice profile " << n << '\t' << m << '\t' ;
 f <<  "  this " << endl;
 f << " pL = " << pL << " L ="  << L << endl
   << " pU = " << pU << " U ="  << U << endl
   << " D = " << D << endl;
 if ( (pL == pU) &&  (U == L) )
   if (pL && L) 
     {f << " matrice profile symetrique " <<endl;
     int i,j,k;
     for (i = 0;i<n;i++) 
       { f << i << " {" << pL[i+1]-pL[i] << "}" <<'\t' ;
       for (k=pL[i];k<pL[i+1];k++)
	 { j=i-(pL[i+1]-k);
	 f << j << " " << L[k] << "; "; 
	 }
       f <<  i  << ":" << D[i] << endl  ;
       }
     }
   else f << " MatriceProfile: pointeur vide " <<endl; 
 else 
   { 
     f << " matrice profile non symetrique " << endl;
     int i,k;
     for (i = 0;i<n;i++) 
       {
	 f << i ;
	 if (pL && L) 
	   {
	     f << " jO=" << i-pL[i+1]+pL[i] << " L= " <<'\t' ;
	     for (k=pL[i];k<pL[i+1];k++)
	       { 
		 f <<  " " << L[k] ; 
	       }
	   }
	 if (D)
	   f  << " D= " << D[i]  << '\t' ;
	 else 
	   f  << " D=0 => 1 ; ";
	 if (pU && U) 
	   {
	     f << " i0=" << i-pU[i+1]+pU[i] << " U= " <<'\t' ;
	     for (k=pU[i];k<pU[i+1];k++)
	       f << " " << U[k] ; 

	   }
	 f << endl;  
       }
    
   }
 return f;
}


template<class R>
void MatriceProfile<R>::cholesky(R eps) const {
  R  *ij , *ii  , *ik , *jk , xii;
  int i,j,k;
  if (L != U) ERREUR(factorise,"matrice non symetrique");
  U = 0; // 
  typefac = FactorizationCholeski;
  D[0] = sqrt(D[0]); 
  ij = L ; // pointeur sur le terme ij de la matrice avec j<i 
  for (i=1;i<n;i++) // boucle sur les lignes 
    { ii = L+pL[i+1]; // pointeur sur le terme fin de la ligne +1 =>  ij < ii;
    xii = D[i] ; 
    for ( ; ij < ii ; ij++) // pour les j la ligne i
      { j = i -(ii - ij); 
      k = Max( j - (pL[j+1]-pL[j]) ,  i-(pL[i+1]-pL[i]) ); 
      ik =  ii - (i - k); 
      jk =  L + pL[j+1] -(j - k); 
      k = j - k ; 
      while ( k-- ) *ij -= *ik++ * *jk++;  
      *ij /=  D[j] ;
      xii -= *ij * *ij ;
      }
    if (xii < eps*Abs(D[i])) 
      ERREUR(cholesky,"pivot (" << i << ")= " << xii << " < " << eps*Abs(D[i]))
	D[i] = sqrt(xii);
    }
}


template<class R>
void MatriceProfile<R>::crout(R eps) const  {
  R  *ij , *ii  , *ik , *jk , xii, *dkk;
  int i,j,k;
  if (L != U) ERREUR(factorise,"matrice non symetrique");
  U = 0; // 
  typefac = FactorizationCrout;
   
  ij = L ; // pointeur sur le terme ij de la matrice avec j<i 
  for (i=1;i<n;i++) // boucle sur les lignes 
    { ii = L+pL[i+1]; // pointeur sur le terme fin de la ligne +1 =>  ij < ii;
    xii = D[i] ; 
    for ( ; ij < ii ; ij++) // pour les j la ligne i
      { j = i -(ii - ij); 
      k = Max( j - (pL[j+1]-pL[j]) ,  i-(pL[i+1]-pL[i]) ); 
      ik =  ii - (i - k); 
      jk =  L + pL[j+1] -(j - k); 
      dkk = D + k;
      k = j - k ; 
      while ( k-- ) *ij -= *ik++ * *jk++ * *dkk++;  
      *ij /=  *dkk ; // k = j ici 
      xii -= *ij * *ij * *dkk;
      }
    if (Abs(xii) <= Max(eps*Abs(D[i]),1.0e-30))
      ERREUR(crout,"pivot (" << i << " )= " << Abs(xii) << " <= " << eps*Abs(D[i]) << " eps = " << eps)
	D[i] = xii;
    }
}


template<class R>
void MatriceProfile<R>::LU(R eps) const  {
  R s,uii;
  int i,j,k;
  if (L == U && ( pL[n]  || pU[n] ) ) ERREUR(LU,"matrice  symetrique");
  if(verbosity>3)
  cout << " -- LU " << endl;
  typefac=FactorizationLU;

  for (i=1;i<n;i++) // boucle sur les sous matrice de rang i 
    { 
      // for L(i,j)  j=j0,i-1
      int j0 = i-(pL[i+1]-pL[i]);
      for ( j = j0; j<i;j++)
        {           
          int k0 = Max(j0,j-(pU[j+1]-pU[j]));
          R *Lik = L + pL[i+1]-i+k0; // lower
          R *Ukj = U + pU[j+1]-j+k0; // upper
          s =0;
          for (k=k0;k<j;k++) // k < j < i ;
	    s += *Lik++ * *Ukj++ ;     // a(i,k)*a(k,j);
	  *Lik -= s;
	  *Lik /= D[j]; //  k == j here
        }
      // for U(j,i) j=0,i-1        
      j0=i-pU[i+1]+pU[i];
      for (j=j0;j<i;j++) 
	{
	  s = 0;
	  int k0 = Max(j0,j-pL[j+1]+pL[j]);
	  R *Ljk = L + pL[j+1]-j+k0;   
	  R *Uki = U + pU[i+1]-i+k0;   
	  for (k=k0  ;k<j;k++)    // 
	    s +=  *Ljk++ * *Uki++ ;
	  *Uki -= s;  // k = j here 
	}
      // for D (i,i) in last because we need L(i,k) and U(k,i) for k<j
      int k0 = i-Min(pL[i+1]-pL[i],pU[i+1]-pU[i]);
      R *Lik = L + pL[i+1]-i+k0; // lower
      R *Uki = U + pU[i+1]-i+k0; // upper
      s =0;
      for (k=k0;k<i;k++) // k < i < i ;
	s += *Lik++ * *Uki++ ;     // a(i,k)*a(k,i);
      // cout << " k0 " << k0 << " i = " << i << " " <<  s << endl;
      uii = D[i] -s;
      
      if (Abs(uii) <= Max(eps*Abs(D[i]),1.0e-30))
	ERREUR(LU,"pivot (" << i << " )= " << Abs(uii) << " <= " << eps*Abs(D[i]) << " eps = " << eps);     
      
      D[i] = uii;
      
    }
}


template<class R>
KN_<R> & operator/=(KN_<R> & x ,const MatriceProfile<R> & a) 
{
  // --------------------------------------------------------------------
  // si La diagonale D n'existe pas alors on suppose 1 dessus (cf crout)
  // --------------------------------------------------------------------
  R * v = &x[0];
  int n = a.n;  
  if (x.n != n ) 
    ERREUR (KN_<R> operator/(MatriceProfile<R>),"  matrice et KN_<R> incompatible");
  const R *ij ,*ii, *ik, *ki;
  R *xk,*xi;
  int i;
  switch (a.typefac) {
  case FactorizationNO:
    if (a.U && a.L) {cerr << "A PROGRAMMER (KN_<R><R>::operator/MatriceProfile)";throw(ErrorExec("exit",2));}
   
    if ( a.U && !a.L ) 
      { // matrice triangulaire superieure
	// cout << " remonter " << (a.D ? "DU" : "U") << endl;
	ki = a.U + a.pU[n]; 
	i = n;
	while ( i-- )
	  { ii = a.U + a.pU[i];
          xi= xk  = v +  i ;
          if (a.D) *xi /= a.D[i];// pour crout ou LU
          while ( ki > ii) 
	    *--xk  -=  *--ki *  *xi ; 
	  }
      }
    else if  ( !a.U && a.L ) 
      { // matrice triangulaire inferieure
	// cout << " descente "  <<( a.D ? "LD" : "L" ) <<endl;
	ii = a.L;
	for (i=0; i<n; i++)
	  { ij = ik = (a.L + a.pL[i+1]) ;  // ii =debut,ij=fin+1 de la ligne 
          xk = v + i;
          R ss = v[i]; 
          while ( ik > ii) 
	    ss -= *--ik * *--xk ; 
          if ( a.D) ss /= a.D[i];// pour crout ou LU
          v[i] = ss ;
          ii = ij;
	  }
      }
    else if (a.D) 
      { // matrice diagonale
	// cout << " diagonal D" <<endl;
	for (i=0;i<n;i++) 
	  v[i]=v[i]/a.D[i];
      }
    break;
  case FactorizationCholeski:  
    //     cout << " FactorizationChosleski" << endl;
    x /= a.ld();
    x /= a.ldt();   
    break;
  case FactorizationCrout:
    //   cout << " FactorizationCrout" << endl;
    x /= a.l();
    x /= a.d();
    x /= a.lt();
    break;
  case FactorizationLU:
    //  cout << " FactorizationLU" << endl;
    x  /= a.l();
    x  /= a.du();
    break;
  default : assert( false );
    /*   default:
	 ERREUR  (operator /=(MatriceProfile," Error unkwon type of Factorization  =" << typefac);
    */
  }
  return x;
}


/*
void Int_L(const Interpolation &Uh,Vecteur &f,const Interpolation &Vh, Vecteur &b, const LOpera & LU,const LOpera & LV)
{  
  int k, i,ii,iii,ii1,j,jj,jjj,jj1;
  throwassert(f.N == Uh.NbOfDF);
  throwassert(b.N == Vh.NbOfDF);
  throwassert(&Vh.Th == &Uh.Th);
  //  throwassert(Uh.N == Vh.N);
  int NU = LU.imax();
  int NV = LV.imax();
  // to stock the value of the basic function
  T_RNM   U(Uh.N,Uh.MaxNbOfDFByElement);
  T_RNMK dU(2,Uh.N,Uh.MaxNbOfDFByElement);
  T_RNM V(Vh.N,Vh.MaxNbOfDFByElement);
  T_RNMK dV(2,Vh.N,Vh.MaxNbOfDFByElement);
 
  T_RN Lu(NV);
  

  for( k=0;k<Uh.NbOfElements;k++)
    { 
      const ElementFini & eu = Uh[k]; 
      const ElementFini & ev = Vh[k]; 
      const  TypeInterpolation & ku = eu; // type Interpolation de element 
      const TypeInterpolation & kv = ev; // type Interpolation de element 
      const QuadratureFormular & FI = QuadratureFormular_T_5;
      const TypeElement &ke = ku;// type de element 
      int n = Uh[k];
      int m = Vh[k];
      const Element & e = Uh.Th[k]; // l'element 
      for ( int npi=0; npi<FI.n; npi++ ) // loop on the integration point
	{
	  Coor P = FI[npi];
	  CoorxCoor J;
	  Lu =0;
	  kv.vFb(P,Vh,k,V);
	  ku.vFb(P,Uh,k,U);
	  R coef;
	  ku.vDFb(P,Uh,k,dU,J,coef); // coef = area of e
	  ku.vDFb(P,Vh,k,dV,J,coef); // coef = area of e
	  coef *=  FI[npi].a;
	  // calcule de LU * f * coef = Lu
	  int nn;
	  for ( i=nn=0; i<n; i++ )
	    for (iii=eu[i],ii=Uh.FirstDFOfNode(iii),ii1=Uh.LastDFOfNode(iii);ii<ii1;ii++,nn++)
	      { 
		T_RN_ Unn(U(nn));
		T_RNM_ dUnn(dU(nn));
		R fii = f[ii];
		for (int l=0;l<LU.k;l++) // for all the linear form 
		  {
		    int ii= LU.i[l];
		    int jj= LU.j[l]; 
		    throwassert(jj==0); 
		    const DOpera &f = LU.v[l];
		    // for all the term of the linear form f
		    for (int ll=0;ll<f.k;ll++)
		      {
			R  v = f.v[ll]*coef*fii;
			int iii = f.i[ll];
			int jjj = f.j[ll];
			if(ii<NV) //all the term  ii>= NV are not used 
			  {
			    if (jjj==0)
			      Lu[ii] += Unn [iii]*v;
			    else
			      Lu[ii] += dUnn(jjj-1,iii)*v;
			  }
		      }
		  }
	      }
	  //  cout << Lu << endl;
	  for (nn=j=0;j<m;j++)
	    for (jjj=ev[j], jj=Vh.FirstDFOfNode(jjj),jj1=Vh.LastDFOfNode(jjj);jj<jj1;jj++,nn++)
	      {
		T_RN_ Vnn(V(nn));
		T_RNM_ dVnn(dV(nn));
		R bb=0;
		//  bb = Lu . Lv 
		for (int l=0;l<LV.k;l++) // for all the term of LV
		  {
		    int ii= LV.i[l];
		    int jj= LV.j[l];
		    throwassert(jj==0);
		    const DOpera &f =LV.v[l];
		     // for all the term of the linear form f
		    for (int ll=0;ll<f.k;ll++)
		      {
			R  v = f.v[ll]*Lu[ii];
			int iii = f.i[ll];
			int jjj = f.j[ll];
			if(ii<NV) //all the term  ii>= NV are not used 
			  {
			    if (jjj==0)
			      bb += Vnn [iii]*v;
			    else
			      bb += dVnn(jjj-1,iii)*v;
			  }
			
		      }
		  }
		b[nn] += bb;
	      }
	}
    }
  
}


*/
#endif

