//#pragma dont_inline on
//#pragma global_optimizer off

extern long verbosity ;
#include <cmath>
#include <cstdlib>
#include "error.hpp"
#include <iostream>
#include <fstream>
#include "RNM.hpp"
#include "rgraph.hpp"
#include "fem.hpp"
#include <set>


namespace Fem3D {

void Triangle::set( Vertex * v0, int i0, int i1, int i2, int l)
{
       lab=l; 
       R3 A=*(vertices[0]=v0+i0);
       R3 B=*(vertices[1]=v0+i1);
       R3 C=*(vertices[2]=v0+i2); 
       area = Norme2(  (( B-A)^(C-A)) ) *0.5;
       throwassert(area>0);
}
 
void Tetraedre::set( Vertex * v0, int i0, int i1, int i2, int i3, int l)
{
       lab=l; 
       R3 &A=*(vertices[0]=v0+i0);
       R3 &B=*(vertices[1]=v0+i1);
       R3 &C=*(vertices[2]=v0+i2); 
       R3 &D=*(vertices[3]=v0+i3);
       volume = ( ((B-A)^(C-A)),(D-A))/6.;
       throwassert(volume>0);
}


void  Tetraedre::print(void) const
{ 
  R3     &D = *(vertices[0]),
         &C = *(vertices[1]),
         &B = *(vertices[2]),
         &A = *(vertices[3]);

  cout << " tetraedre::Hi " <<  this << " points are : "<< endl;
  cout << " (" <<  &A << ") A.x,.. Are " <<  A.x << " " << A.y << " " << A.z<< endl;
  cout << " (" <<  &B << ") B.x,.. Are " <<  B.x << " " << B.y << " " << B.z<< endl;
  cout << " (" <<  &C << ") C.x,.. Are " <<  C.x << " " << C.y << " " << C.z<< endl;
  cout << " (" <<  &D << ") D.x,.. Are " <<  D.x << " " << D.y << " " << D.z<< endl;
}



R3 Tetraedre::H( int i )  const
{ 
  assert( i>=0 ); assert( i<4 );

  // si i est impair, on inverse c et b pour avoir les
  // indices dans l'ordre direct
  R3     &D = *(vertices[  i      ]),
         &C = *(vertices[ (i+1 + (i%2))%4 ]),
         &B = *(vertices[ (i+2 - (i%2))%4 ]),
         &A = *(vertices[ (i+3)%4 ]);

  // check we took the coordinates in the right order
  assert  ( ((B-A)^(C-A),(D-A)) >0 );

  // On cherche HD, avec H dans le plan ABC et HD perpendiculaire a ce plan
  R3 h = ( B-A )^( C-A ); // normal a la face ABC
  R norm = Norme2(h);;
  h /= norm; // h est unitaire
  h /= (6.* volume/norm); // on divise par la longueur de la hauteur
  return h;
}
 

  
 
void  Mesh::BoundingBox(R3 &Pmin,R3 &Pmax) const 
{
    throwassert(nv);
    Pmin=Pmax=vertices[0];
    for (int i=0;i<nv;i++)
        { 
            const R3 & P=vertices[i];
            Pmin.x = Min(Pmin.x,P.x);
            Pmin.y = Min(Pmin.y,P.y);
            Pmin.z = Min(Pmin.z,P.z);
            Pmax.x = Max(Pmax.x,P.x);
            Pmax.y = Max(Pmax.y,P.y);
            Pmin.z = Max(Pmin.z,P.z);
        } 
  //  cout << " Bounding Box = " << Pmin <<" " <<  Pmax << endl;       
}


void Mesh::read(const char * filename)
{ // read the mesh on the file of name filename
  ////////////////////////////////////////////////////////////////////////
  // Z Modifications 28/10/2003 pour traiter le cas 3d Alain Perronnet
  //                            Ajout de ndim=2 => Triangles
  //                                         =3 => Tetraedres
  // comme seul parametre de la premiere ligne du fichier
  // De plus le label d'un sommet s'il est un point est -no du Point
  //                                       un sommet sur une surface est no de surface
  // cela pour differencier le label d'un point de celui de surface
  // si un sommet appartient a plusieurs surfaces, c'est la derniere
  // qui est prise en compte => pour eviter tout doute introduire un point
  // L'idee est de traiter ainsi plusieurs types de conditions aux limites
  ////////////////////////////////////////////////////////////////////////
    dim=3;
    ntetra=0;
#ifdef USE_GIBBS
    TriangleConteningVertex =0;
    BoundaryAdjacencesHead=0;
    BoundaryAdjacencesLink=0;
    BoundaryEdgeHeadLink=0;
    TheAdjacencesLink =0;
#endif // USE_GIBBS
    volume=0.;
    
    int i,i0,i1,i2,i3,ir;
    ifstream f(filename);
    if (!f) {
         cerr << "Erreur ouverture du fichier " << filename << endl;
         throw(ErrorExec("exit",1));}
   // throwassert(f);
   if(verbosity)
    cout << " Read On file \"" <<filename<<"\""<<  endl;

   f >> dim ;
   if(verbosity)
     cout << "   Dimension of Mesh " << dim << endl;
   assert (dim==3);

   f >> nv >> ntetra >> nbt ;
   if(verbosity)
     cout << "   Nb of Vertex " << nv << " " << " Nb of Tetraedres " 
          << ntetra << " Nb of boundary triangles " << nbt <<  endl;
   throwassert(f.good() && ntetra && nv && nbt) ;

   //Allocation dynamique des tableaux du maillage
   vertices  = new Vertex[nv];
   tetraedres = new Tetraedre[ntetra];
   btriangles = new Triangle [nbt];
   throwassert(btriangles && vertices && tetraedres);
   if(verbosity)
     cout << "   Allocated space : vertices,tetraedres,btriangles are :" 
	     << vertices << ", "
	     << tetraedres << ", "
	     << btriangles << endl;
   volume=0.;

   // read vertices
   for (i=0;i<nv;i++)    // Lecture X Y Label de chacun des vertices
     f >> vertices[i],throwassert(f.good());

   // read tetraedres
   for (i=0;i<ntetra;i++) { 
     f >> i0 >> i1 >> i2 >> i3 >> ir;
     throwassert(f.good() && i0>0 && i0<=nv && i1>0 && i1<=nv && i2>0 && i2<=nv && i3>0 && i3<=nv);
     tetraedres[i].set(vertices,i0-1,i1-1,i2-1,i3-1,ir); 
     volume += tetraedres[i].volume; }
   
   // read boundary triangles
   for (i=0;i<nbt;i++) {
     f >> i0 >> i1 >> i2 >> ir;
     throwassert(f.good() && i0>0 && i0<=nv && i1>0 && i1<=nv && i2>0 && i2<=nv );
     btriangles[i].set(vertices,i0-1,i1-1,i2-1,ir); }
    
   if(verbosity)
    cout << "   End of read: volume on mesh = " << volume <<endl;  

   // ConsAdjacence();

}

     
Mesh::~Mesh()
 {
   SHOWVERB(cout << "   -- delete mesh " << this << endl);
   delete [] vertices;
 // delete [] adj;
#ifdef USE_GIBBS
   delete [] TheAdjacencesLink;
   delete [] BoundaryEdgeHeadLink;
   delete [] BoundaryAdjacencesHead;
   delete [] BoundaryAdjacencesLink;
    delete []  TriangleConteningVertex;
#endif // USE_GIBBS
 }

     
} // namespace
