#ifndef FEM2_H_
#define FEM2_H_
#include <string> 
#include "RefCounter.hpp"

// some usefull functions ------------------------------------------------------------------
template<class K> class KN_;

const  double Pi =  3.14159265358979323846264338328;

template<class T> inline T Min (const T &a,const T &b){return a < b ? a : b;}
template<class T> inline T Max (const T &a,const T & b){return a > b ? a : b;}
template<class T> inline T Abs (const T &a){return a <0 ? -a : a;}

template<class T> inline void Exchange (T& a,T& b) {T c=a;a=b;b=c;}
template<class T> inline T Max (const T &a,const T & b,const T & c){return Max(Max(a,b),c);}
template<class T> inline T Min (const T &a,const T & b,const T & c){return Min(Min(a,b),c);}

// definition R  ---------------------------------------------------------------------------
typedef double R;

#define USE_GIBBS

namespace Fem3D  //=========================================================================
{

typedef KN_<R>   RN_;
typedef KN<R>    RN;
typedef KNM_<R>  RNM_;
typedef KNM<R>   RNM;
typedef KNMK_<R> RNMK_;
typedef KNMK<R>  RNMK;


// The class R3  ---------------------------------------------------------------------------
class R3 {
  friend ostream& operator <<(ostream& f, const R3 & P )
  { f << P.x << ' ' << P.y << ' ' << P.z   ; return f; }
  friend istream& operator >>(istream& f,  R3 & P)
  { f >> P.x >>  P.y >>  P.z  ; return f; }

public:  
  R  x,y,z;
 
  R3 () :x( 0. ),y( 0. ),z(0.) {};
  R3 (R a,R b,R c):x(a),y(b),z(c)  {}
  void set (R a,R b,R c){x=a;y=b;z=c;}
//  R3 (R2 P2):x(P2.x ),y(P2.y),z(0)  {}
  R3 (const R3 & A,const R3 & B) :x(B.x-A.x ),y( B.y-A.y),z(B.z-A.z)  {}
//  R3 & operator=(R2 P2) {x=P2.x;y=P2.y;z=0;return *this;}
  R3   operator+(R3 P)const   {return R3(x+P.x,y+P.y,z+P.z);}
  R3 & operator+=(R3 P)  {x += P.x;y += P.y;z += P.z;return *this;}
  R3   operator-(R3 P)const   {return R3(x-P.x,y-P.y,z-P.z);}
  R3 & operator-=(R3 P) {x -= P.x;y -= P.y;z -= P.z;return *this;}
  R3   operator-()const  {return R3(-x,-y,-z);}
  R3   operator+()const  {return *this;}
  R    operator,(R3 P)const  {return  x*P.x+y*P.y+z*P.z;} //produit scalaire
  R3  operator^(R3 P)const {return R3(y*P.z-z*P.y ,P.x*z-x*P.z, x*P.y-y*P.x);} //produit vectoriel
  R3   operator*(R c)const {return R3(x*c,y*c,z*c);}
  R3   operator/(R c)const {return R3(x/c,y/c,z/c);}
  R3   operator*=(R c)  {x *= c;y *= c; z *=c; return *this;}
  R3   operator/=(R c)  {x /= c;y /= c; z /=c; return *this;}
  R  &  operator[](int i){ return (&x)[i];}
  friend R3 operator*(R c,R3 P) {return P*c;}
  friend R3 operator/(R c,R3 P) {return P/c;}
};

inline R Norme2_2(const R3 & A){ return (A,A);}
inline R Norme2(const R3 & A){ return sqrt((A,A));}
                                
// reference number for the physics  ------------------------------------------------------------
class Label
{
public: 
  int lab;
  Label(int r=0):lab(r){}
  int  operator!() const{return !lab;} 
  bool operator<(const Label & r)  const {return lab < r.lab; } 
  bool operator==(const Label & r) const {return lab == r.lab;} 
  bool operator>(const Label & r)  const {return lab > r.lab; } 
};

inline ostream& operator <<( ostream& f, const Label & r  )
  { f <<  r.lab ; return f; }
inline istream& operator >>( istream& f, Label & r  )
  { f >>  r.lab ; return f; }
  

class Vertex : public R3, public Label //----------------------------------------------------
{
  friend class Mesh;
public:
  Vertex() : R3(), Label(){};
  Vertex( R3 P, int r=0 ) : R3(P), Label(r) {}
};

inline ostream& operator << ( ostream& f, const Vertex & v )
 { f << v.x << ' ' << v.y << ' ' << v.z << ( Label ) v  ; return f; }
inline istream& operator >> ( istream& f,  Vertex & v )
  { f >>  v.x >>  v.y >>  v.z >> ( Label& ) v  ; return f; }


class Triangle: public Label  //------------------------------------------------------------
{
private:
  Vertex *vertices[3]; // an array of 3 pointers to vertex
public:
  R area;

  Triangle(){};        // constructor empty for array

  Triangle( Vertex * v0, int i0, int i1, int i2, int l) : Label(l) 
     { set(v0,i0,i1,i2,l); }

  Vertex & operator[](int i) const // to see triangle as an array of vertex
       { assert( i>=0 ); assert(i<3); return *vertices[i];} 
       
  Vertex * & operator()(int i)     // to see triangle as an array of vertex
       { assert( i>=0 ); assert(i<3); return vertices[i];} 

  void set( Vertex * v0, int i0, int i1, int i2, int l);

         
  void Draw(double shink=1) const;
  void Fill(int color) const;
  void Draw(int edge,double shink=1) const;

  void Renum(  Vertex  *v0, long * r )
  {
  for ( int i=0; i<3; i++ ) vertices[ i ]=v0+r[ vertices[ i ]-v0 ];
  }

};



class Tetraedre: public Label  //------------------------------------------------------------
{
private:
  Vertex *vertices[4]; // an array of 4 pointers to vertex
public:
  R volume;

  Tetraedre(){};        // constructor empty for array

  Tetraedre( Vertex * v0, int i0, int i1, int i2, int i3, int l) : Label(l) 
     { set(v0,i0,i1,i2,i3,l); }

  Vertex & operator[](int i) const // to see triangle as an array of vertex
       { //       cout << "operator[] i is :" <<i<<endl;
       assert(i>=0); assert(i<4); return *vertices[i];} 
       
  Vertex * operator()(int i)     // to see triangle as an array of vertex
       { assert(i>=0); assert(i<4); return vertices[i];} 

  void set( Vertex * v0, int i0, int i1, int i2, int i3, int l);

  R3 H( int i )  const; // donne la hauteur arrivant au sommet i
  void print(void) const ;  // for debugging purpose
      
  void Renum( Vertex  *v0, long * r )
  {  for (int i=0; i<4; i++) 
       vertices[i]=v0+r[vertices[i]-v0];}
};


class Mesh: public RefCounter  //----------------------------------------------------------
{
public:
/*
 * par rapport a 2d :
 *
	 nv -> nv
	 nt -> ntetra
	 triangles -> tetraedres
	 neb -> nbt
	 bedges -> btriangles

 */
  R    volume; //total du maillage
  int  dim;    //2 ou 3

  int  nv;     //nombre de vertices
  Vertex       * vertices;     //tableau des vertices  du maillage
  int  ntetra; // nbre de tetraedre
  Tetraedre    * tetraedres; 
  int  nbt;
  Triangle     * btriangles;   // boundary triangles
  
  //Les constructeurs
  Mesh( const char * filename ) { read(filename); } // read on a file
  Mesh( const string s ) { read(s.c_str()); }

 ~Mesh();

  Tetraedre & operator[](int i) const  //le tetraedre i du Mesh
    { throwassert(i>=0 && i<ntetra); return tetraedres[i]; }

  Vertex   & operator()(int i) const  //le vertex i du Mesh
    { throwassert(i>=0 && i<nv); return vertices[i]; }

  Triangle & getBoundaryTriangle( int i ) {
  assert( i>=0 ); assert ( i<nbt ); return btriangles[i]; }

  int getBoundaryTriangleVerticeIndice( int t, int i ) {
  assert( t>=0 ); assert ( t<nbt );
  assert( i>=0 ); assert ( i<3 );
  return number( btriangles[t][i] );
  }

  int number(const Tetraedre & t) const {return &t - tetraedres;}
  int number(const Tetraedre * t) const {return  t - tetraedres;}
  int number(const Vertex & v)   const {return &v - vertices;}
  int number(const Vertex * v)   const {return  v - vertices;}
  int operator()(int it, int j) const {
	  assert(it>=0); assert(it<ntetra);
	  assert(j>=0); assert(j<4);
	  int n = number(tetraedres[it][j]);
	  assert( n>=0 ); assert( n<nv );
	  return n;
	  }
                 // Numero du vertex j du tetraedre it

  void BoundingBox(R3 & Pmin,R3 &Pmax) const;

  void InitDraw() const ;
//  void Draw(int init=2,bool fill=false) const;
//  void DrawBoundary() const;

#ifdef USE_GIBBS
  int renum();
  int gibbsv (long* ptvoi,long* vois,long* lvois,long* w,long* v);
#endif // USE_GIBBS

  void destroy() {RefCounter::destroy();}
  void MakeQuadTree();

private:
  void read(const char * filename); 
#ifdef USE_GIBBS
  void BuildBoundaryAdjacences();

  // to construct the adj triangle               
  int *TheAdjacencesLink;
  int *BoundaryEdgeHeadLink;
  int *BoundaryAdjacencesHead;
  int *BoundaryAdjacencesLink; 
  int *TriangleConteningVertex;               
#endif // USE_GIBBS
};

#if 0
//  2 routines to compute the caracteristics
int WalkInTriangle( const Mesh & Th, int it, double *lambda,
                    const KN_<R> & U, const KN_<R> & V, R & dt );
int WalkInTriangle( const Mesh & Th, int it, double *lambda,
		    R u, R v, R & dt ); 
                   
int Walk( const Mesh & Th,int & it, R *l,
          const KN_<R> & U,const KN_<R> & V, R dt ) ;
#endif

         
template<class T>
void  HeapSort(T *c,long n)
{
  long l,j,r,i;
  T crit;
  c--; // on decale de 1 pour que le tableau commence a 1
  if( n <= 1) return;
  l = n/2 + 1;
  r = n;
  while (1)
  { // label 2
    if(l <= 1 )
    { // label 20
      crit = c[r];
      c[r--] = c[1];
      if ( r == 1 ) { c[1]=crit; return;}
    }
    else
      crit = c[--l]; 
    j=l;
    while (1)
    { // label 4
      i=j;
      j=2*j;
      if  (j>r) {c[i]=crit;break;} // L8 -> G2
      if ((j<r) && (c[j] < c[j+1])) j++; // L5
      if (crit < c[j])
	c[i]=c[j]; // L6+1 G4
      else
	{c[i]=crit;break;} //L8 -> G2
    }
  }
}

}  //Fin namespace Fem3D =========================================================

#endif
